import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import javax.microedition.io.Connector;
import com.nttdocomo.ui.*;

public final class GBACanvas extends Canvas {

	private static final String SOFTKEY_ROOT_STR   = "MENU";
	private static final String SOFTKEY_GAME_STR   = "ްщ";
	private static final String SOFTKEY_ROM_STR    = "̕ҏW";
	private static final String SOFTKEY_RESET_STR  = "RESET";

	// 萔
	private static final int MODE_GAME = 1;
	private static final int MODE_MENU_ROM = 2;

	/** y[WSDROM. */
	private static final int NUM_SDROM = 12;

	/** ROM̒ */
	private static final int SIZE_ROM_NAME = 30;

	// CX^Xϐ
	private GBACore core;

	public static int[] fb = new int[240*160];

	// SDROMt@CXg
	private String[] sdRomnm_ = new String[NUM_SDROM];

	// N̓j[\
	private int mode_ = MODE_MENU_ROM;
	// SubMenu
	private int selectSubMenuNo_ = 0;

	private String fps = "";
	private final int frameskip = 5;
	private int frame_count;
	private long last_time;

	private Image allImg = null;
	private Graphics allG = null;

	private final Font DEFAULT_FONT;
	private final int HEIGHT_LINE;
	private final int BASE_LINE;

	{
		DEFAULT_FONT = Font.getDefaultFont();
		// s = tHg + ̗] + ̗]
		HEIGHT_LINE = DEFAULT_FONT.getHeight() + 1 + 2 * 2;
		// x[X = ] + tHgAscent
		BASE_LINE = 3 + DEFAULT_FONT.getAscent();
	}

	public GBACanvas() {
		setSoftLabel(Frame.SOFT_KEY_1, SOFTKEY_GAME_STR);
		setSoftLabel(Frame.SOFT_KEY_2, SOFTKEY_ROM_STR);
		setBackground(Graphics.getColorOfName(Graphics.BLACK));

		// XNb`pbhǍ
		readSdromNameSP();

		allImg = Image.createImage(getWidth(), getHeight());
		allG = allImg.getGraphics();
		allG.setColor(Graphics.getColorOfName(Graphics.BLACK));
		allG.fillRect(0, 0, getWidth(), getHeight());

		core = new GBACore(this);
	}

	public final boolean startFrame() {
		frame_count++;
		return frame_count % frameskip == 0;
	}

	Graphics g = getGraphics();

	public void endFrame() {
		if (frame_count >= 60) {
			long current = System.currentTimeMillis();
			fps = (int)((current - last_time) * (60/1000)) + "fps";
			last_time = current;
			frame_count = 0;
		}
		paint(g);
	}

	public final short readJoyPad() {
		int key, Input;

		key = getKeypadState();
		Input = 0x03FF;

		if ((key & (1 << Display.KEY_3))     == 0x00008) Input &= ~0x0001; // A
		if ((key & (1 << Display.KEY_2))     == 0x00004) Input &= ~0x0002; // B
		if ((key & (1 << Display.KEY_5))     == 0x00020) Input &= ~0x0004; // Select
		if ((key & (1 << Display.KEY_6))     == 0x00040) Input &= ~0x0008; // Start
		if ((key & (1 << Display.KEY_RIGHT)) == 0x40000) Input &= ~0x0010; // 
		if ((key & (1 << Display.KEY_LEFT))  == 0x10000) Input &= ~0x0020; // 
		if ((key & (1 << Display.KEY_UP))    == 0x20000) Input &= ~0x0040; // 
		if ((key & (1 << Display.KEY_DOWN))  == 0x80000) Input &= ~0x0080; // 
		if ((key & (1 << Display.KEY_9))     == 0x00200) Input &= ~0x0100; // R
		if ((key & (1 << Display.KEY_8))     == 0x00100) Input &= ~0x0200; // L

		return (short) Input;
	}

	public void setMessage(final String s) {
		g.lock();
		g.setColor(Graphics.getColorOfName(Graphics.BLACK));
		g.fillRect(0, 224, 240, 16);
		g.setColor(Graphics.getColorOfName(Graphics.WHITE));
		g.drawString(s, 4, 224 + BASE_LINE);
		g.unlock(true);
	}

	public void processEvent(final int type, final int param) {
		if (type != Display.KEY_PRESSED_EVENT) return;
		switch (mode_) {
			case MODE_GAME :
				switch (param) {
					case Display.KEY_SOFT1 :
						core.request(core.REQUEST_SLEEP, 0);
						setSoftLabel(Frame.SOFT_KEY_1, SOFTKEY_GAME_STR);
						setSoftLabel(Frame.SOFT_KEY_2, SOFTKEY_ROM_STR);
						mode_ = MODE_MENU_ROM;
						repaint();
						break;

					case Display.KEY_SOFT2 :
						core.request(core.REQUEST_SLEEP, 0);
						if (confirm("Zbg܂H"))
							core.request(core.REQUEST_RESET, 0);
						else
							core.request(core.REQUEST_START, 0);
						break;

					default :
				}
				break;
			case MODE_MENU_ROM :
				switch (param) {
					case Display.KEY_SOFT1 :
						core.request(core.REQUEST_START, 0);
						setSoftLabel(Frame.SOFT_KEY_1, SOFTKEY_ROOT_STR);
						setSoftLabel(Frame.SOFT_KEY_2, SOFTKEY_RESET_STR);
						mode_ = MODE_GAME;
						repaint();
						break;

					case Display.KEY_SOFT2 :
						imeOn(sdRomnm_[selectSubMenuNo_].trim(), TextBox.DISPLAY_ANY, TextBox.KANA);
						break;

					case Display.KEY_SELECT :
						loadRom();
						break;

					case Display.KEY_UP :
						if (selectSubMenuNo_ > 0)
							selectSubMenuNo_--;
						else
							selectSubMenuNo_ = 11;
						repaint();
						break;

					case Display.KEY_DOWN :
						if (selectSubMenuNo_ < NUM_SDROM - 1)
							selectSubMenuNo_++;
						else
							selectSubMenuNo_ = 0;
						repaint();
						break;

					default :
				}
				break;
		}
	}

	public void processIMEEvent(int type, String text) {
		if (type != IME_COMMITTED) return;

		// ROM̕ҏW
		text = formatSpace(text, SIZE_ROM_NAME);
		sdRomnm_[selectSubMenuNo_] = text;
		writeSdromNameSP(selectSubMenuNo_, text);
		repaint();
	}

	public void paint(final Graphics g) {
		g.lock();
		if (mode_ == MODE_GAME) {
			allG.setRGBPixels(0,40,240,160,fb,0);

			// FPS
			if (fps != null) {
				allG.setColor(Graphics.getColorOfName(Graphics.BLACK));
				allG.fillRect(0, 224, 240, 16);
				allG.setColor(Graphics.getColorOfName(Graphics.WHITE));
				allG.drawString(fps, 4, 224 + BASE_LINE);
			}

			g.drawImage(allImg, 0, 0);
		}
		else {
			g.drawImage(allImg, 0, 0);

			int x = 50;
			int wid = DEFAULT_FONT.stringWidth("                    "); //p20

			for (int i=0; i<sdRomnm_.length; i++) {
				if (DEFAULT_FONT.stringWidth(sdRomnm_[i].trim()) > wid) {
					wid = DEFAULT_FONT.stringWidth(sdRomnm_[i].trim());
					x = 50 - ((wid - 120) / 2);
				}
			}
			drawChs(sdRomnm_, x, 10, wid + 20);
			drawcTriangle(x + 5, HEIGHT_LINE * (selectSubMenuNo_ + 1));
		}
		g.unlock(true);
	}

	// j[\GA̕`
	public void drawMsg(int xp, int yp, int wid, int hgt) {
		int[] x = {xp, xp+2, xp+wid-2, xp+wid, xp+wid, xp+wid-2, xp+2, xp, xp};
		int[] y = {yp+2, yp, yp, yp+2, yp+hgt-2, yp+hgt, yp+hgt, yp+hgt-2, yp+2};
		int[] x2 = {xp-1, xp+2, xp+wid-2, xp+wid+1, xp+wid+1, xp+wid-2, xp+2, xp-1, xp-1};
		int[] y2 = {yp+2, yp-1, yp-1, yp+2, yp+hgt-2, yp+hgt+1, yp+hgt+1, yp+hgt-2, yp+2};

		g.setColor(g.getColorOfRGB(0x33, 0x33, 0x77, 0xCF));
		g.fillRect(xp, yp, wid, hgt);

		g.setColor(g.getColorOfRGB(0x88, 0x88, 0x88));
		g.drawPolyline(x, y, 9);
		g.setColor(g.getColorOfRGB(0xCC, 0xCC, 0xCC));
		g.drawPolyline(x2, y2, 9);
	}

	// J[\̕`
	public void drawcTriangle(int xp, int yp) {
		int[] x = {xp, xp+5, xp, xp};
		int[] y = {yp, yp+5, yp+10, yp};

		g.setColor(g.getColorOfName(g.WHITE));
		g.fillPolygon(x, y, 4);
	}

	// j[̕`
	public void drawChs(String[] sChois, int x, int y, int wid) {
		g.setFont(DEFAULT_FONT);
		int myp = y + HEIGHT_LINE;

		drawMsg(x, y, wid, (HEIGHT_LINE * sChois.length) + 7);

		g.setColor(g.getColorOfName(g.WHITE));
		for(int i=0; i<sChois.length; i++) g.drawString(sChois[i], x + 17, myp + (i * HEIGHT_LINE));
	}


	private void createSdfiles() {
		//_~[SDt@C쐬
		SDbinding sd = new SDbinding();
		for(int i = 1; i <= NUM_SDROM; i++){
			sd.createFile(formatNumber(i, 3) + ".GBA");
		}
	}

	private void createBiosfile() {
		//_~[BIOS쐬
		SDbinding sd = new SDbinding();
		sd.createFile("BIOS.BIN");
	}

	private void deleteSdfiles(final int page) {
		//_~[SDt@C폜
	}

	private void loadRom() {
		setMessage("ROMt@Cǂݍݒ");
		SDbinding sd = null;
		InputStream is = null;
		try {
			//GBACore̍ċN
			core.request(core.REQUEST_STOP, 0);
			core = new GBACore(this);
			System.gc();

			String fnm = formatNumber(selectSubMenuNo_ + 1, 3) + ".GBA";
			sd = new SDbinding();
			is = sd.openInputStream(fnm);
			if (is != null) {
				int c = core.loadROM(is, sd.getFileSize());
				if (c == 0) {
					setSoftLabel(Frame.SOFT_KEY_1, SOFTKEY_ROOT_STR);
					setSoftLabel(Frame.SOFT_KEY_2, SOFTKEY_RESET_STR);
					mode_ = MODE_GAME;
					core.reset();
					core.start();
					core.request(core.REQUEST_START, 0);
				}
				else {
					if (c == -1)
						alert("GBAt@Cł͂܂B");
					else if (c == -2)
						alert("BIOSt@CsłB");
				}
			}
			else {
				if (confirm("ROMt@Cǂݍ߂܂B\n_~[t@C쐬܂H\n("
					     + 1 + "y[W" + NUM_SDROM + "t@C)")) {
					createSdfiles();
					createBiosfile();
				}
			}
		} catch (Exception e) {
			alert("err loadRom: " + e);
		} catch (Error err) {
			alert("err loadRomError: " + err);
			throw err;
		} finally {
			is = null;
			sd.close();
			sd = null;
			//SDbinding.closeFileEntry();
		}
	}

	private void readSdromNameSP() {
		DataInputStream in = null;
		try {
			in = Connector.openDataInputStream(
				"scratchpad:///0;pos="
					+ 0
					+ ",length=" + SIZE_ROM_NAME * NUM_SDROM);
			byte[] name = new byte[SIZE_ROM_NAME];
			for (int i = 0; i < NUM_SDROM; i++) {
				in.read(name);
				if (name[0] == 0) {
					sdRomnm_[i] = "";
				} else {
					sdRomnm_[i] = new String(name);
				}
			}
		} catch (Exception e) {
			alert("err readSdromNameSP: " + e);
		} finally {
			try {
				in.close();
			} catch (Exception ex) {}
			in = null;
		}
	}

	private void writeSdromNameSP(final int sdno, final String romnm) {
		// SDRomF30
		DataOutputStream out = null;
		try {
			out = Connector.openDataOutputStream(
				"scratchpad:///0;pos="
					+ (SIZE_ROM_NAME * sdno)
					+ ",length=" + SIZE_ROM_NAME);
			out.write(formatSpace(romnm, SIZE_ROM_NAME).getBytes());
			out.flush();
		} catch (Exception e) {
			alert("err writeSdromNameSP: " + e);
		} finally {
			try {
				out.close();
			} catch (Exception ex) {}
			out = null;
		}
	}

	/////////////////////////////////
	// Utility Method
	/////////////////////////////////

	public void alert(final String s) {
		while (getKeypadState() != 0) {
			try {
				Thread.sleep(10L);
			} catch (Exception ex) {}
		}
		Dialog dialog = new Dialog(Dialog.DIALOG_ERROR, "G[I");
		dialog.setText(s);
		dialog.show();
	}

	public void infomsg(final String s) {
		while (getKeypadState() != 0) {
			try {
				Thread.sleep(10L);
			} catch (Exception ex) {}
		}
		Dialog dialog = new Dialog(Dialog.DIALOG_INFO , "");
		dialog.setText(s);
		dialog.show();
	}

	private boolean confirm(final String s) {
		while (getKeypadState() != 0) {
			try {
				Thread.sleep(10L);
			} catch (Exception ex) {}
		}
		Dialog dialog = new Dialog(Dialog.DIALOG_YESNO, "mF");
		dialog.setText(s);
		boolean ret = (dialog.show() == Dialog.BUTTON_YES);
		return ret;
	}

	private String formatNumber(final int num, final int size) {
		String str = String.valueOf(num);
		while (str.length() < size) {
			str = "0" + str;
		}
		return str;
	}

	private String formatSpace(final String str, final int size) {
		String ret;
		byte[] inbyte = str.getBytes();
		if (inbyte.length > size) {
			ret = new String(inbyte, 0, size);
		} else {
			ret = str;
		}
		while (ret.getBytes().length < size) {
			ret = ret.concat(" ");
		}
		return ret;
	}

}
